@@ -5,7 +5,9 @@ import enum
 import functools
 import inspect
 import socket
+import sys
 import warnings
+from asyncio import AbstractEventLoopPolicy
 from textwrap import dedent
 from typing import (
     Any,
@@ -19,20 +21,23 @@ from typing import (
     Literal,
     Optional,
     Set,
+    Type,
     TypeVar,
     Union,
     overload,
 )
 
 import pytest
-from _pytest.mark.structures import get_unpacked_marks
 from pytest import (
+    Class,
     Collector,
     Config,
     FixtureRequest,
     Function,
     Item,
     Metafunc,
+    Module,
+    Package,
     Parser,
     PytestCollectionWarning,
     PytestDeprecationWarning,
@@ -184,12 +189,6 @@ def pytest_configure(config: Config) -> None:
         "mark the test as a coroutine, it will be "
         "run using an asyncio event loop",
     )
-    config.addinivalue_line(
-        "markers",
-        "asyncio_event_loop: "
-        "Provides an asyncio event loop in the scope of the marked test "
-        "class or module",
-    )
 
 
 @pytest.hookimpl(tryfirst=True)
@@ -206,11 +205,6 @@ def _preprocess_async_fixtures(
     config = collector.config
     asyncio_mode = _get_asyncio_mode(config)
     fixturemanager = config.pluginmanager.get_plugin("funcmanage")
-    event_loop_fixture_id = "event_loop"
-    for node, mark in collector.iter_markers_with_node("asyncio_event_loop"):
-        event_loop_fixture_id = node.stash.get(_event_loop_fixture_id, None)
-        if event_loop_fixture_id:
-            break
     for fixtures in fixturemanager._arg2fixturedefs.values():
         for fixturedef in fixtures:
             func = fixturedef.func
@@ -222,6 +216,14 @@ def _preprocess_async_fixtures(
                 # Ignore async fixtures without explicit asyncio mark in strict mode
                 # This applies to pytest_trio fixtures, for example
                 continue
+            scope = fixturedef.scope
+            if scope == "function":
+                event_loop_fixture_id = "event_loop"
+            else:
+                event_loop_node = _retrieve_scope_root(collector, scope)
+                event_loop_fixture_id = event_loop_node.stash.get(
+                    _event_loop_fixture_id, None
+                )
             _make_asyncio_fixture_function(func)
             function_signature = inspect.signature(func)
             if "event_loop" in function_signature.parameters:
@@ -364,18 +366,19 @@ class PytestAsyncioFunction(Function):
     """Base class for all test functions managed by pytest-asyncio."""
 
     @classmethod
-    def substitute(cls, item: Function, /) -> Function:
+    def item_subclass_for(
+        cls, item: Function, /
+    ) -> Union[Type["PytestAsyncioFunction"], None]:
         """
-        Returns a PytestAsyncioFunction if there is an implementation that can handle
-        the specified function item.
+        Returns a subclass of PytestAsyncioFunction if there is a specialized subclass
+        for the specified function item.
 
-        If no implementation of PytestAsyncioFunction can handle the specified item,
-        the item is returned unchanged.
+        Return None if no specialized subclass exists for the specified item.
         """
         for subclass in cls.__subclasses__():
             if subclass._can_substitute(item):
-                return subclass._from_function(item)
-        return item
+                return subclass
+        return None
 
     @classmethod
     def _from_function(cls, function: Function, /) -> Function:
@@ -383,6 +386,7 @@ class PytestAsyncioFunction(Function):
         Instantiates this specific PytestAsyncioFunction type from the specified
         Function item.
         """
+        assert function.get_closest_marker("asyncio")
         subclass_instance = cls.from_parent(
             function.parent,
             name=function.name,
@@ -392,6 +396,7 @@ class PytestAsyncioFunction(Function):
             keywords=function.keywords,
             originalname=function.originalname,
         )
+        subclass_instance.own_markers.extend(function.own_markers)
         subclassed_function_signature = inspect.signature(subclass_instance.obj)
         if "event_loop" in subclassed_function_signature.parameters:
             subclass_instance.warn(
@@ -418,11 +423,10 @@ class Coroutine(PytestAsyncioFunction):
         return asyncio.iscoroutinefunction(func)
 
     def runtest(self) -> None:
-        if self.get_closest_marker("asyncio"):
-            self.obj = wrap_in_sync(
-                # https://github.com/pytest-dev/pytest-asyncio/issues/596
-                self.obj,  # type: ignore[has-type]
-            )
+        self.obj = wrap_in_sync(
+            # https://github.com/pytest-dev/pytest-asyncio/issues/596
+            self.obj,  # type: ignore[has-type]
+        )
         super().runtest()
 
 
@@ -462,11 +466,10 @@ class AsyncStaticMethod(PytestAsyncioFunction):
         )
 
     def runtest(self) -> None:
-        if self.get_closest_marker("asyncio"):
-            self.obj = wrap_in_sync(
-                # https://github.com/pytest-dev/pytest-asyncio/issues/596
-                self.obj,  # type: ignore[has-type]
-            )
+        self.obj = wrap_in_sync(
+            # https://github.com/pytest-dev/pytest-asyncio/issues/596
+            self.obj,  # type: ignore[has-type]
+        )
         super().runtest()
 
 
@@ -484,10 +487,9 @@ class AsyncHypothesisTest(PytestAsyncioFunction):
         ) and asyncio.iscoroutinefunction(func.hypothesis.inner_test)
 
     def runtest(self) -> None:
-        if self.get_closest_marker("asyncio"):
-            self.obj.hypothesis.inner_test = wrap_in_sync(
-                self.obj.hypothesis.inner_test,
-            )
+        self.obj.hypothesis.inner_test = wrap_in_sync(
+            self.obj.hypothesis.inner_test,
+        )
         super().runtest()
 
 
@@ -534,89 +536,119 @@ def pytest_pycollect_makeitem_convert_async_functions_to_subclass(
     for node in node_iterator:
         updated_item = node
         if isinstance(node, Function):
-            updated_item = PytestAsyncioFunction.substitute(node)
+            specialized_item_class = PytestAsyncioFunction.item_subclass_for(node)
+            if specialized_item_class:
+                if _get_asyncio_mode(
+                    node.config
+                ) == Mode.AUTO and not node.get_closest_marker("asyncio"):
+                    node.add_marker("asyncio")
+                if node.get_closest_marker("asyncio"):
+                    updated_item = specialized_item_class._from_function(node)
         updated_node_collection.append(updated_item)
-
     hook_result.force_result(updated_node_collection)
 
 
 _event_loop_fixture_id = StashKey[str]
+_fixture_scope_by_collector_type = {
+    Class: "class",
+    # Package is a subclass of module and the dict is used in isinstance checks
+    # Therefore, the order matters and Package needs to appear before Module
+    Package: "package",
+    Module: "module",
+    Session: "session",
+}
 
 
 @pytest.hookimpl
 def pytest_collectstart(collector: pytest.Collector):
-    if not isinstance(collector, (pytest.Class, pytest.Module)):
+    try:
+        collector_scope = next(
+            scope
+            for cls, scope in _fixture_scope_by_collector_type.items()
+            if isinstance(collector, cls)
+        )
+    except StopIteration:
         return
-    # pytest.Collector.own_markers is empty at this point,
-    # so we rely on _pytest.mark.structures.get_unpacked_marks
-    marks = get_unpacked_marks(collector.obj, consider_mro=True)
-    for mark in marks:
-        if not mark.name == "asyncio_event_loop":
-            continue
-        event_loop_policy = mark.kwargs.get("policy", asyncio.get_event_loop_policy())
-        policy_params = (
-            event_loop_policy
-            if isinstance(event_loop_policy, Iterable)
-            else (event_loop_policy,)
-        )
-
-        # There seem to be issues when a fixture is shadowed by another fixture
-        # and both differ in their params.
-        # https://github.com/pytest-dev/pytest/issues/2043
-        # https://github.com/pytest-dev/pytest/issues/11350
-        # As such, we assign a unique name for each event_loop fixture.
-        # The fixture name is stored in the collector's Stash, so it can
-        # be injected when setting up the test
-        event_loop_fixture_id = f"{collector.nodeid}::<event_loop>"
+    # Session is not a PyCollector type, so it doesn't have a corresponding
+    # "obj" attribute to attach a dynamic fixture function to.
+    # However, there's only one session per pytest run, so there's no need to
+    # create the fixture dynamically. We can simply define a session-scoped
+    # event loop fixture once in the plugin code.
+    if collector_scope == "session":
+        event_loop_fixture_id = _session_event_loop.__name__
         collector.stash[_event_loop_fixture_id] = event_loop_fixture_id
+        return
+    # There seem to be issues when a fixture is shadowed by another fixture
+    # and both differ in their params.
+    # https://github.com/pytest-dev/pytest/issues/2043
+    # https://github.com/pytest-dev/pytest/issues/11350
+    # As such, we assign a unique name for each event_loop fixture.
+    # The fixture name is stored in the collector's Stash, so it can
+    # be injected when setting up the test
+    event_loop_fixture_id = f"{collector.nodeid}::<event_loop>"
+    collector.stash[_event_loop_fixture_id] = event_loop_fixture_id
 
-        @pytest.fixture(
-            scope="class" if isinstance(collector, pytest.Class) else "module",
-            name=event_loop_fixture_id,
-            params=policy_params,
-            ids=tuple(type(policy).__name__ for policy in policy_params),
-        )
-        def scoped_event_loop(
-            *args,  # Function needs to accept "cls" when collected by pytest.Class
-            request,
-        ) -> Iterator[asyncio.AbstractEventLoop]:
-            new_loop_policy = request.param
-            old_loop_policy = asyncio.get_event_loop_policy()
-            old_loop = asyncio.get_event_loop()
-            asyncio.set_event_loop_policy(new_loop_policy)
+    @pytest.fixture(
+        scope=collector_scope,
+        name=event_loop_fixture_id,
+    )
+    def scoped_event_loop(
+        *args,  # Function needs to accept "cls" when collected by pytest.Class
+        event_loop_policy,
+    ) -> Iterator[asyncio.AbstractEventLoop]:
+        new_loop_policy = event_loop_policy
+        with _temporary_event_loop_policy(new_loop_policy):
             loop = asyncio.new_event_loop()
             asyncio.set_event_loop(loop)
             yield loop
             loop.close()
-            asyncio.set_event_loop_policy(old_loop_policy)
-            asyncio.set_event_loop(old_loop)
 
-        # @pytest.fixture does not register the fixture anywhere, so pytest doesn't
-        # know it exists. We work around this by attaching the fixture function to the
-        # collected Python class, where it will be picked up by pytest.Class.collect()
-        # or pytest.Module.collect(), respectively
-        collector.obj.__pytest_asyncio_scoped_event_loop = scoped_event_loop
-        break
+    # @pytest.fixture does not register the fixture anywhere, so pytest doesn't
+    # know it exists. We work around this by attaching the fixture function to the
+    # collected Python class, where it will be picked up by pytest.Class.collect()
+    # or pytest.Module.collect(), respectively
+    collector.obj.__pytest_asyncio_scoped_event_loop = scoped_event_loop
+    # When collector is a package, collector.obj is the package's __init__.py.
+    # pytest doesn't seem to collect fixtures in __init__.py.
+    # Using parsefactories to collect fixtures in __init__.py their baseid will end
+    # with "__init__.py", thus limiting the scope of the fixture to the init module.
+    # Therefore, we tell the pluginmanager explicitly to collect the fixtures
+    # in the init module, but strip "__init__.py" from the baseid
+    # Possibly related to https://github.com/pytest-dev/pytest/issues/4085
+    if isinstance(collector, Package):
+        fixturemanager = collector.config.pluginmanager.get_plugin("funcmanage")
+        package_node_id = _removesuffix(collector.nodeid, "__init__.py")
+        fixturemanager.parsefactories(collector.obj, nodeid=package_node_id)
 
 
-def pytest_collection_modifyitems(
-    session: Session, config: Config, items: List[Item]
-) -> None:
-    """
-    Marks collected async test items as `asyncio` tests.
+def _removesuffix(s: str, suffix: str) -> str:
+    if sys.version_info < (3, 9):
+        return s[: -len(suffix)]
+    return s.removesuffix(suffix)
 
-    The mark is only applied in `AUTO` mode. It is applied to:
 
-      - coroutines and async generators
-      - Hypothesis tests wrapping coroutines
-      - staticmethods wrapping coroutines
-
-    """
-    if _get_asyncio_mode(config) != Mode.AUTO:
-        return
-    for item in items:
-        if isinstance(item, PytestAsyncioFunction):
-            item.add_marker("asyncio")
+@contextlib.contextmanager
+def _temporary_event_loop_policy(policy: AbstractEventLoopPolicy) -> Iterator[None]:
+    old_loop_policy = asyncio.get_event_loop_policy()
+    try:
+        old_loop = asyncio.get_event_loop()
+    except RuntimeError:
+        old_loop = None
+    asyncio.set_event_loop_policy(policy)
+    try:
+        yield
+    finally:
+        asyncio.set_event_loop_policy(old_loop_policy)
+        # When a test uses both a scoped event loop and the event_loop fixture,
+        # the "_provide_clean_event_loop" finalizer of the event_loop fixture
+        # will already have installed a fresh event loop, in order to shield
+        # subsequent tests from side-effects. We close this loop before restoring
+        # the old loop to avoid ResourceWarnings.
+        try:
+            asyncio.get_event_loop().close()
+        except RuntimeError:
+            pass
+        asyncio.set_event_loop(old_loop)
 
 
 _REDEFINED_EVENT_LOOP_FIXTURE_WARNING = dedent(
@@ -625,40 +657,49 @@ _REDEFINED_EVENT_LOOP_FIXTURE_WARNING = dedent(
     %s:%d
     Replacing the event_loop fixture with a custom implementation is deprecated
     and will lead to errors in the future.
-    If you want to request an asyncio event loop with a class or module scope,
-    please attach the asyncio_event_loop mark to the respective class or module.
+    If you want to request an asyncio event loop with a scope other than function
+    scope, use the "scope" argument to the asyncio mark when marking the tests.
+    If you want to return different types of event loops, use the event_loop_policy
+    fixture.
     """
 )
 
 
 @pytest.hookimpl(tryfirst=True)
 def pytest_generate_tests(metafunc: Metafunc) -> None:
-    for event_loop_provider_node, _ in metafunc.definition.iter_markers_with_node(
-        "asyncio_event_loop"
-    ):
-        event_loop_fixture_id = event_loop_provider_node.stash.get(
-            _event_loop_fixture_id, None
-        )
-        if event_loop_fixture_id:
-            # This specific fixture name may already be in metafunc.argnames, if this
-            # test indirectly depends on the fixture. For example, this is the case
-            # when the test depends on an async fixture, both of which share the same
-            # asyncio_event_loop mark.
-            if event_loop_fixture_id in metafunc.fixturenames:
-                continue
-            fixturemanager = metafunc.config.pluginmanager.get_plugin("funcmanage")
-            if "event_loop" in metafunc.fixturenames:
-                raise MultipleEventLoopsRequestedError(
-                    _MULTIPLE_LOOPS_REQUESTED_ERROR
-                    % (metafunc.definition.nodeid, event_loop_provider_node.nodeid),
-                )
-            # Add the scoped event loop fixture to Metafunc's list of fixture names and
-            # fixturedefs and leave the actual parametrization to pytest
-            metafunc.fixturenames.insert(0, event_loop_fixture_id)
-            metafunc._arg2fixturedefs[
-                event_loop_fixture_id
-            ] = fixturemanager._arg2fixturedefs[event_loop_fixture_id]
-            break
+    marker = metafunc.definition.get_closest_marker("asyncio")
+    if not marker:
+        return
+    scope = marker.kwargs.get("scope", "function")
+    if scope == "function":
+        return
+    event_loop_node = _retrieve_scope_root(metafunc.definition, scope)
+    event_loop_fixture_id = event_loop_node.stash.get(_event_loop_fixture_id, None)
+
+    if event_loop_fixture_id:
+        # This specific fixture name may already be in metafunc.argnames, if this
+        # test indirectly depends on the fixture. For example, this is the case
+        # when the test depends on an async fixture, both of which share the same
+        # event loop fixture mark.
+        if event_loop_fixture_id in metafunc.fixturenames:
+            return
+        fixturemanager = metafunc.config.pluginmanager.get_plugin("funcmanage")
+        if "event_loop" in metafunc.fixturenames:
+            raise MultipleEventLoopsRequestedError(
+                _MULTIPLE_LOOPS_REQUESTED_ERROR.format(
+                    test_name=metafunc.definition.nodeid,
+                    scope=scope,
+                    scoped_loop_node=event_loop_node.nodeid,
+                ),
+            )
+        # Add the scoped event loop fixture to Metafunc's list of fixture names and
+        # fixturedefs and leave the actual parametrization to pytest
+        # The fixture needs to be appended to avoid messing up the fixture evaluation
+        # order
+        metafunc.fixturenames.append(event_loop_fixture_id)
+        metafunc._arg2fixturedefs[
+            event_loop_fixture_id
+        ] = fixturemanager._arg2fixturedefs[event_loop_fixture_id]
 
 
 @pytest.hookimpl(hookwrapper=True)
@@ -675,6 +716,7 @@ def pytest_fixture_setup(
         _add_finalizers(
             fixturedef,
             _close_event_loop,
+            _restore_event_loop_policy(asyncio.get_event_loop_policy()),
             _provide_clean_event_loop,
         )
         outcome = yield
@@ -749,6 +791,23 @@ def _close_event_loop() -> None:
         loop.close()
 
 
+def _restore_event_loop_policy(previous_policy) -> Callable[[], None]:
+    def _restore_policy():
+        # Close any event loop associated with the old loop policy
+        # to avoid ResourceWarnings in the _provide_clean_event_loop finalizer
+        try:
+            with warnings.catch_warnings():
+                warnings.simplefilter("ignore", DeprecationWarning)
+                loop = previous_policy.get_event_loop()
+        except RuntimeError:
+            loop = None
+        if loop:
+            loop.close()
+        asyncio.set_event_loop_policy(previous_policy)
+
+    return _restore_policy
+
+
 def _provide_clean_event_loop() -> None:
     # At this point, the event loop for the current thread is closed.
     # When a user calls asyncio.get_event_loop(), they will get a closed loop.
@@ -820,11 +879,11 @@ def wrap_in_sync(
 _MULTIPLE_LOOPS_REQUESTED_ERROR = dedent(
     """\
         Multiple asyncio event loops with different scopes have been requested
-        by %s. The test explicitly requests the event_loop fixture, while another
-        event loop is provided by %s.
+        by {test_name}. The test explicitly requests the event_loop fixture, while
+        another event loop with {scope} scope is provided by {scoped_loop_node}.
         Remove "event_loop" from the requested fixture in your test to run the test
-        in a larger-scoped event loop or remove the "asyncio_event_loop" mark to run
-        the test in a function-scoped event loop.
+        in a {scope}-scoped event loop or remove the scope argument from the "asyncio"
+        mark to run the test in a function-scoped event loop.
     """
 )
 
@@ -833,16 +892,15 @@ def pytest_runtest_setup(item: pytest.Item) -> None:
     marker = item.get_closest_marker("asyncio")
     if marker is None:
         return
-    event_loop_fixture_id = "event_loop"
-    for node, mark in item.iter_markers_with_node("asyncio_event_loop"):
-        event_loop_fixture_id = node.stash.get(_event_loop_fixture_id, None)
-        if event_loop_fixture_id:
-            break
+    scope = marker.kwargs.get("scope", "function")
+    if scope != "function":
+        parent_node = _retrieve_scope_root(item, scope)
+        event_loop_fixture_id = parent_node.stash[_event_loop_fixture_id]
+    else:
+        event_loop_fixture_id = "event_loop"
     fixturenames = item.fixturenames  # type: ignore[attr-defined]
-    # inject an event loop fixture for all async tests
-    if "event_loop" in fixturenames:
-        fixturenames.remove("event_loop")
-    fixturenames.insert(0, event_loop_fixture_id)
+    if event_loop_fixture_id not in fixturenames:
+        fixturenames.append(event_loop_fixture_id)
     obj = getattr(item, "obj", None)
     if not getattr(obj, "hypothesis", False) and getattr(
         obj, "is_hypothesis_test", False
@@ -853,9 +911,29 @@ def pytest_runtest_setup(item: pytest.Item) -> None:
         )
 
 
+def _retrieve_scope_root(item: Union[Collector, Item], scope: str) -> Collector:
+    node_type_by_scope = {
+        "class": Class,
+        "module": Module,
+        "package": Package,
+        "session": Session,
+    }
+    scope_root_type = node_type_by_scope[scope]
+    for node in reversed(item.listchain()):
+        if isinstance(node, scope_root_type):
+            return node
+    error_message = (
+        f"{item.name} is marked to be run in an event loop with scope {scope}, "
+        f"but is not part of any {scope}."
+    )
+    raise pytest.UsageError(error_message)
+
+
 @pytest.fixture
 def event_loop(request: FixtureRequest) -> Iterator[asyncio.AbstractEventLoop]:
     """Create an instance of the default event loop for each test case."""
+    new_loop_policy = request.getfixturevalue(event_loop_policy.__name__)
+    asyncio.set_event_loop_policy(new_loop_policy)
     loop = asyncio.get_event_loop_policy().new_event_loop()
     # Add a magic value to the event loop, so pytest-asyncio can determine if the
     # event_loop fixture was overridden. Other implementations of event_loop don't
@@ -867,6 +945,29 @@ def event_loop(request: FixtureRequest) -> Iterator[asyncio.AbstractEventLoop]:
     loop.close()
 
 
+@pytest.fixture(scope="session")
+def _session_event_loop(
+    request: FixtureRequest, event_loop_policy: AbstractEventLoopPolicy
+) -> Iterator[asyncio.AbstractEventLoop]:
+    new_loop_policy = event_loop_policy
+    with _temporary_event_loop_policy(new_loop_policy):
+        loop = asyncio.new_event_loop()
+        asyncio.set_event_loop(loop)
+        yield loop
+        loop.close()
+
+
+@pytest.fixture(scope="session", autouse=True)
+def event_loop_policy() -> AbstractEventLoopPolicy:
+    """Return an instance of the policy used to create asyncio event loops."""
+    return asyncio.get_event_loop_policy()
+
+
+def is_async_test(item: Item) -> bool:
+    """Returns whether a test item is a pytest-asyncio test"""
+    return isinstance(item, PytestAsyncioFunction)
+
+
 def _unused_port(socket_type: int) -> int:
     """Find an unused localhost port from 1024-65535 and return it."""
     with contextlib.closing(socket.socket(type=socket_type)) as sock:
